<?php

namespace Drupal\cloudflare_stream\Plugin\Field\FieldType;

use Drupal\Component\Render\FormattableMarkup;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\TypedData\DataDefinition;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\TypedData\DataDefinitionInterface;
use Drupal\Core\TypedData\TraversableTypedDataInterface;
use Drupal\Core\TypedData\TypedDataInterface;
use Drupal\file\FileUsage\FileUsageInterface;
use Drupal\file\Plugin\Field\FieldType\FileItem;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\GuzzleException;

/**
 * Plugin implementation of the 'cloudflareVideo' field type.
 *
 * @FieldType(
 *   id = "cloudflarevideo",
 *   label = @Translation("Cloudflare Video"),
 *   description = @Translation("This field stores the ID of an cloudflare video."),
 *   category = @Translation("Cloudflare"),
 *   default_widget = "cloudflarevideo_default",
 *   default_formatter = "cloudflarevideo_default",
 *   list_class = "\Drupal\file\Plugin\Field\FieldType\FileFieldItemList",
 *   constraints = {"ReferenceAccess" = {}, "FileValidation" = {}}
 * )
 */
class CloudflareVideoItem extends FileItem {

  /*
   * Http timeout time.
   */
  protected $timeout;

  /**
   * The configuration factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The HTTP client.
   *
   * @var \GuzzleHttp\Client
   */
  protected $httpClient;

  /**
   * The file system.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * The Messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The file usage service.
   *
   * @var \Drupal\file\FileUsage\FileUsageInterface
   */
  protected $fileUsage;

  /**
   * {@inheritdoc}
   */
  public function __construct(
    DataDefinitionInterface $definition,
    ConfigFactoryInterface $config_factory,
    EntityTypeManagerInterface $entity_type_manager,
    ClientInterface $http_client,
    FileSystemInterface $file_system,
    $name = NULL,
    TypedDataInterface $parent = NULL,
    MessengerInterface $messenger = NULL,
    FileUsageInterface $file_usage = NULL
  ) {
    parent::__construct($definition, $name, $parent);
    $this->configFactory = $config_factory->get('cloudflare_stream.settings');
    $this->entityTypeManager = $entity_type_manager;
    $this->httpClient = $http_client;
    $this->fileSystem = $file_system;
    $this->messenger = $messenger;
    $this->timeout = 600;
    $this->fileUsage = $file_usage;
  }

  /**
   * {@inheritdoc}
   */
  public static function createInstance($definition, $name = NULL, TraversableTypedDataInterface $parent = NULL) {
    $configFactory = \Drupal::service('config.factory');
    $entityTypeManager = \Drupal::service('entity_type.manager');
    $httpClient = \Drupal::service('http_client');
    $fileSystem = \Drupal::service('file_system');
    $messenger = \Drupal::service('messenger');
    $fileUsage = \Drupal::service('file.usage');
    return new static(
      $definition,
      $configFactory,
      $entityTypeManager,
      $httpClient,
      $fileSystem,
      $name,
      $parent,
      $messenger,
      $fileUsage
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultFieldSettings() {
    $fieldSettings = parent::defaultFieldSettings();
    $fieldSettings['file_extensions'] = 'mp4 avi mov';

    return $fieldSettings;
  }

  /**
   * {@inheritdoc}
   */
  public static function schema(FieldStorageDefinitionInterface $field_definition) {
    $schema = parent::schema($field_definition);
    $schema['columns']['cloudflareStreamVideoID'] = [
      'type' => 'text',
      'length' => 512,
      'not null' => FALSE,
    ];
    $schema['columns']['thumbnail'] = [
      'type' => 'text',
      'length' => 512,
      'not null' => FALSE,
    ];

    return $schema;
  }

  /**
   * {@inheritdoc}
   */
  public static function propertyDefinitions(FieldStorageDefinitionInterface $field_definition) {
    $properties = parent::propertyDefinitions($field_definition);

    $properties['cloudflareStreamVideoID'] = DataDefinition::create('string')
      ->setLabel(t('Cloudflare Video ID'));
    $properties['thumbnail'] = DataDefinition::create('string')
      ->setLabel(t('Cloudflare Video thumbnail'));
    return $properties;
  }

  /*
   * This function returns the header array for api request.
   */
  public function getHeaders() {
    return [
      'X-Auth-Email' => $this->getEmail(),
      'X-Auth-Key' => $this->getApiKey(),
    ];
  }

  /*
   * This function returns the email id from cloudflare stream configuration.
   */
  public function getEmail() {
    return $this->configFactory->get('email');
  }

  /*
   * This function returns the api_key from cloudflare stream configuration.
   */
  public function getApiKey() {
    return $this->configFactory->get('api_key');
  }

  /*
   * This function returns the zone ID from cloudflare stream configuration.
   */
  public function getZoneID() {
    return $this->configFactory->get('zone_id');
  }

  /*
   * This function returns the account ID from cloudflare stream configuration.
   */
  public function getAccountID() {
    return $this->configFactory->get('account_id');
  }

  /**
   * {@inheritdoc}
   */
  public function preSave() {
    $nid = $this->getEntity()->id();
    if (is_null($this->getValue()['cloudflareStreamVideoID'])) {
      $fileID = $this->get('target_id')->getValue();
      $file = $this->entityTypeManager->getStorage('file')->load($fileID);
      if (isset($file)) {
        $fileRealPath = $this->fileSystem->realpath($file->getFileUri());
        $filename = $file->getFilename();
        try {
          $request = $this->httpClient->post("https://api.cloudflare.com/client/v4/accounts/{$this->getAccountID()}/media", [
            'timeout' => $this->getTimeout(),
            'headers' => $this->getHeaders(),
            'multipart' => [
              [
                'name' => 'file',
                'contents' => file_get_contents($fileRealPath),
                'filename' => $filename,
              ],
            ],
          ]);

          $response = Json::decode($request->getBody()->getContents());
          $videoID = $response['result']['uid'];
          $oldVideoID = $this->getOldVideoID($nid, $this->name);
          if (isset($oldVideoID)) {
            $this->deleteCloudflareVideo($oldVideoID, $this->getHeaders());
          }
          $this->setValue(
            [
              'cloudflareStreamVideoID' => $videoID,
              'target_id' => $this->getValue()['target_id'],
              'thumbnail' => $response['result']['thumbnail'],
            ]
          );
          $this->fileUsage->add($file, 'cloudflare_stream', 'cloudflarevideo', \Drupal::currentUser()->id());
          $file->setPermanent();
          $file->save();
        } catch (GuzzleException $error) {
          // Get the original response.
          $response = $error->getResponse();
          // Get the info returned from the remote server.
          $response_info = Json::decode($response->getBody()->getContents());
          // Using FormattableMarkup allows for the use of <pre/> tags,
          // giving a more readable log item.
          $message = new FormattableMarkup('CloudflareVideoItem(Upload): API connection error. Error details are as follows:<pre>@response</pre>', ['@response' => print_r($response_info['errors'], TRUE)]);
          // Log the error.
          watchdog_exception('Cloudflare Stream', $error, $message);
          $this->messenger->addError($message);
        }
      }
    }
    else {
      $node = $this->entityTypeManager->getStorage('node')->load($nid);
      $fieldName = $this->getParent()->getName();
      $newFieldData = $this->getEntity()->get($fieldName)->getValue();
      $originalFieldData = $node->get($fieldName)->getValue();
      $videoIds = array_filter($this->getRemovedID($originalFieldData, $newFieldData), function($element) {
        if ($element !== NULL) {
          return $element;
        }
      });

      foreach ($videoIds as $videoId) {
        $this->deleteCloudflareVideo($videoId, $this->getHeaders());
      }
    }
  }

  /**
   * Delete previously uploaded video.
   * @param $id
   * @param $headers
   */
  protected function deleteCloudflareVideo($id, $headers) {
    try {
      $headers['Content-Type'] = 'application/json';
      $this->httpClient->delete("https://api.cloudflare.com/client/v4/zones/{$this->getZoneID()}/media/$id", [
        'headers' => $headers
      ]);
    }
    catch (GuzzleException $error) {
      // Get the original response.
      $response = $error->getResponse();
      // Get the info returned from the remote server.
      $response_info = Json::decode($response->getBody()->getContents());
      // Using FormattableMarkup allows for the use of <pre/> tags,
      // giving a more readable log item.
      $message = new FormattableMarkup('CloudflareVideoItem(Delete): API connection error. Error details are as follows:<pre>@response</pre>', ['@response' => print_r($response_info['errors'], TRUE)]);
      // Log the error.
      watchdog_exception('Cloudflare Stream', $error, $message);
      $this->messenger->addError($message);
    }
  }

  /*
   * This function returns the timeout time.
   */
  protected function getTimeout() {
    return $this->timeout;
  }

  /*
   * This function sets the timeout time.
   */
  public function setTimeout($time) {
    $this->timeout = $time;
  }

  /*
   * This function returns the old video ID.
   */
  public function getOldVideoID($nid, $delta) {
    $oldVideoID = NULL;
    if (!is_null($nid)) {
      $node = $this->entityTypeManager->getStorage('node')->load($nid);
      $fieldName = $this->getParent()->getName();
      $oldVideoID = $node->get($fieldName)->getValue()[$delta]['cloudflareStreamVideoID'] ? $node->get($fieldName)->getValue()[$delta]['cloudflareStreamVideoID'] : NULL;
    }
    return $oldVideoID;
  }

  /**
   * {@inheritdoc}
   */
  public function delete() {
    $this->deleteCloudflareVideo($this->getValue()['cloudflareStreamVideoID'], $this->getHeaders());
  }

  /**
   * @param $originalFieldData
   * @param $newFieldData
   * @return array
   */
  public function getRemovedID($originalFieldData, $newFieldData) {
    $removedId = [];
    foreach ($originalFieldData as $key => $originalFieldDatum) {
      foreach ($newFieldData as $newFieldDatum) {
        if ($originalFieldDatum['cloudflareStreamVideoID'] === $newFieldDatum['cloudflareStreamVideoID']) {
          $removedId[$key] = NULL;
          break;
        }
        $removedId[$key] = $originalFieldDatum['cloudflareStreamVideoID'];
      }
    }

    return $removedId;
  }

}
